(** Differentiability, Taylor expansions *)

Require Export BigO.
Open Local Scope nat_scope.
Open Local Scope R_scope.

Section Diff.

Variable xmin xmax:R.
Hypothesis xminLtxmax: xmin < xmax.

Definition my_proj (X:({x| Betw xmin xmax x}*R)): R*R := (projT1 (fst X), snd X).

(** Differentiability *)

Variable partial_derive_firstvar: (R * R -> R) -> (R * R -> R).
Variable partial_derive_secondvar: (R * R -> R) -> (R * R -> R).

Fixpoint partial_derive_firstvar_n (m : nat) (f : R * R -> R) : R * R -> R :=
  match m with
  | 0 => f
  | S m' => partial_derive_firstvar (partial_derive_firstvar_n m' f)
  end.

Fixpoint partial_derive_secondvar_n (k : nat) (f : R * R -> R) : R * R -> R :=
  match k with
  | 0 => f
  | S k' => partial_derive_secondvar (partial_derive_secondvar_n k' f)
  end.

Definition partial_derive (m k : nat) (f : R * R -> R) : R * R -> R :=
  partial_derive_firstvar_n m (partial_derive_secondvar_n k f).

(** Taylor expansions *)

Definition differential (p : nat) (f : R * R -> R) (X dX : R * R) : R :=
  sum_f_R0 
    (fun m =>
      C p m *
      partial_derive m (p - m)%nat f X *
      fst dX ^ m * snd dX ^ (p - m)%nat)
    p.

Definition DL_pol (n : nat) (f : R * R -> R) (X dX : R * R) : R :=
  sum_f_R0
    (fun p =>
      differential p f X dX / INR (fact p))
    n.

(** Definition of being sufficiently regular by the Taylor expansions *)

Definition is_sufficiently_regular_n m f :=
  Ou (fun X:({x| Betw xmin xmax x}*R) => 
     subtract2_fun
      (fun dX => f (add2 (my_proj X) dX))
      (fun dX => DL_pol (m - 1)%nat f (my_proj X) dX))
    (fun dX => norm_l2 dX ^ m).

(** A few lemmas needed by the consistancy (many computations inside) *)

Lemma OuP_dX_to_eps_Betw: forall P:R*R -> Prop, forall f g eps,
 { C:R | 0 < C /\ (forall (X:({x| Betw xmin xmax x}*R)) dX, norm_l2 (eps (my_proj X) dX) <= C * norm_l2 dX)} ->
 { C:R | 0 < C /\ (forall (X:({x| Betw xmin xmax x}*R)) dX, Rabs (g (eps (my_proj X) dX)) <= C * Rabs (g dX))} ->
 (forall X dX, P dX -> P (eps X dX)) ->
 OuP P (fun (X:({x| Betw xmin xmax x}*R)) dX => f (my_proj X) dX) g -> 
 OuP P (fun (X:({x| Betw xmin xmax x}*R)) dX => f (my_proj X) (eps (my_proj X) dX)) g.
intros P f g eps (C1,(HC1,H1)) (C2,(HC2,H2)) H (alp,(C,(Hc,(Halp,H4)))). 
exists (alp/C1); exists (C2*C).
split.
apply Rmult_lt_0_compat; auto with real.
split.
apply Rmult_lt_0_compat; auto.
intros.
apply Rle_trans with (C* (C2*Rabs (g dX)));[idtac|right; ring].
apply Rle_trans with (C* Rabs (g (eps (my_proj X) dX))).
apply H4.
apply Rle_lt_trans with (C1*norm_l2 dX).
apply H1.
apply Rmult_lt_reg_l with (/C1); auto with real.
apply Rle_lt_trans with (norm_l2 dX).
right; field; auto with real.
apply Rlt_le_trans with (1:= H0); right; unfold Rdiv; ring.
now apply H.
apply Rmult_le_compat_l.
auto with real.
apply H2.
Qed.




Lemma DL_discretization_eps_aux:   forall f : R * R -> R, forall eps: R * R -> R * R -> R * R,
  (is_sufficiently_regular_n 4 f) ->
  { C : R |
     0 < C /\ (forall (X:({x| Betw xmin xmax x}*R)) (dX : R * R), norm_l2 (eps (my_proj X) dX) <= C * norm_l2 dX)} ->
  Ou ((fun (X:({x| Betw xmin xmax x}*R)) dX => f (add2 (my_proj X) (eps (my_proj X) dX)) - 2*f (my_proj X) + f (subtract2 (my_proj X) (eps (my_proj X) dX)) 
           - (fst (eps (my_proj X) dX))^2 * (partial_derive_firstvar_n 2 f (my_proj X))
           - (snd (eps (my_proj X) dX))^2 * (partial_derive_secondvar_n 2 f (my_proj X))
           - 2*(fst (eps (my_proj X) dX))*(snd (eps (my_proj X) dX)) * partial_derive_firstvar (partial_derive_secondvar f) (my_proj X)
            ))
     (fun dX => (norm_l2 dX)^4).
intros f eps H (C', (J1,J2)).
(* c'est exactement la somme de 2 DL *)
apply OuP_eq_trans_l with (fun (X:({x| Betw xmin xmax x}*R)) => add2_fun
    (fun dX => f (add2 (my_proj X) (eps (my_proj X) dX))  - DL_pol (4 - 1)%nat f (my_proj X) (eps (my_proj X) dX))
    (fun dX => f (add2 (my_proj X) (scalar_multiply2 (-1) (eps (my_proj X) dX))) 
           - DL_pol (4 - 1)%nat f (my_proj X) (scalar_multiply2 (-1) (eps (my_proj X) dX)))).
intros; unfold add2_fun; unfold DL_pol,scalar_multiply2, subtract2, add2 ; simpl.
unfold differential; simpl.
unfold partial_derive; simpl.
unfold C; simpl.
replace (snd a + -1 * snd (eps (my_proj a) dX)) with (snd a - snd (eps (my_proj a) dX)) by ring.
replace (projT1 (fst a) + -1 * fst (eps (my_proj a) dX)) with (projT1 (fst a) - fst (eps (my_proj a) dX)) by ring.
field.
apply OuP_is_EV_plus.
(* DL normal *)
apply OuP_dX_to_eps_Betw with (eps:=eps) (f:= fun X dX => f (X+dX)%type - DL_pol (4-1) f X dX).
exists C'; split; auto with real.
exists (C'*C'*C'*C'); split.
now repeat apply Rmult_lt_0_compat.
intros; repeat rewrite Rabs_right.
apply Rle_trans with ((C' * norm_l2 dX)^4).
apply pow_incr.
split; auto.
apply norm_l2_pos.
right; unfold pow; ring.
apply Rle_ge; apply pow_le; apply norm_l2_pos.
apply Rle_ge; apply pow_le; apply norm_l2_pos.
intros; auto.
now apply OuP_eq_trans_l with (2 := H).
(* DL avec -eps X dX *)
apply OuP_dX_to_eps_Betw with (eps:=fun X dX => scalar_multiply2 (-1) (eps X dX)) 
     (f:= fun X dX => f (X+dX)%type - DL_pol (4-1) f X dX).
3: easy.
exists C'; split; auto with real.
intros; rewrite norm_l2_scalar_mult.
rewrite Rabs_Ropp; rewrite Rabs_R1; rewrite Rmult_1_l.
apply J2.
exists (C'*C'*C'*C'); split.
now repeat apply Rmult_lt_0_compat.
intros; repeat rewrite Rabs_right.
rewrite norm_l2_scalar_mult.
rewrite Rabs_Ropp; rewrite Rabs_R1; rewrite Rmult_1_l.
apply Rle_trans with ((C' * norm_l2 dX)^4).
apply pow_incr.
split; auto.
apply norm_l2_pos.
right; unfold pow; ring.
apply Rle_ge; apply pow_le; apply norm_l2_pos.
apply Rle_ge; apply pow_le; apply norm_l2_pos.
now apply OuP_eq_trans_l with (2 := H).
Qed.



Lemma DL_discretization_eps_dx:   forall f : R * R -> R, 
  (is_sufficiently_regular_n 4 f) ->
  Ou ((fun (X:({x| Betw xmin xmax x}*R)) dX => f (add2 (my_proj X) (fst dX, 0)) - 2*f (my_proj X) + f (subtract2 (my_proj X) (fst dX,0)) 
           - (fst dX)^2 * (partial_derive_firstvar_n 2 f (my_proj X))))
     (fun dX => (fst dX)^4).
intros.
apply OuP_eq_trans_r with (fun dX => norm_l2 (fst dX,0) ^4).
intros; rewrite norm_l2_proj1.
rewrite RPow_abs.
apply Rabs_right.
apply Rle_ge; replace ((fst dX) ^ 4) with (Rsqr (Rsqr (fst dX))); auto with real.
unfold Rsqr, pow; ring.
apply OuP_eq_trans_l with (fun (X:({x| Betw xmin xmax x}*R)) dX => (f ((my_proj X) + (fst dX, 0))%type - 2 * f (my_proj X) + f ((my_proj X) - (fst dX, 0))%type 
   - (fst (fst dX,0)) ^ 2 * partial_derive_firstvar_n 2 f (my_proj X)
   - (snd (fst dX,0))^2 * (partial_derive_secondvar_n 2 f (my_proj X))
   - 2*(fst (fst dX,0))*(snd (fst dX,0)) * partial_derive_firstvar (partial_derive_secondvar f) (my_proj X))).
intros; simpl; ring.
apply OuP_implies_OuP1 with 
  (f:=(fun (X:({x| Betw xmin xmax x}*R)) dX => f (add2 (my_proj X) dX) - 2*f (my_proj X) + f (subtract2 (my_proj X) dX) 
           - (fst dX)^2 * (partial_derive_firstvar_n 2 f (my_proj X))
           - (snd dX)^2 * (partial_derive_secondvar_n 2 f (my_proj X))
           - 2*(fst dX)*(snd dX) * partial_derive_firstvar (partial_derive_secondvar f) (my_proj X))) 
  (g:= fun dX => (norm_l2 dX)^4).
intros; auto.
apply DL_discretization_eps_aux with (f:=f) (eps := fun X dX => dX); auto.
exists 1; split; auto with real.
Qed.


Lemma DL_discretization_eps_dt:   forall f : R * R -> R, 
  (is_sufficiently_regular_n 4 f) ->
  Ou ((fun (X:({x| Betw xmin xmax x}*R)) dX => f (add2 (my_proj X) (0, snd dX)) - 2*f (my_proj X) + f (subtract2 (my_proj X) (0, snd dX)) 
           - (snd dX)^2 * (partial_derive_secondvar_n 2 f (my_proj X))))
     (fun dX => (snd dX)^4).
intros.
apply OuP_eq_trans_r with (fun dX => norm_l2 (0, snd dX) ^4).
intros; rewrite norm_l2_proj2.
rewrite RPow_abs.
apply Rabs_right.
apply Rle_ge; replace ((snd dX) ^ 4) with (Rsqr (Rsqr (snd dX))); auto with real.
unfold Rsqr, pow; ring.
apply OuP_eq_trans_l with (fun (X:({x| Betw xmin xmax x}*R)) dX => (f ((my_proj X) + (0, snd dX))%type - 2 * f (my_proj X) + f ((my_proj X) - (0,snd dX))%type 
   - (fst (0,snd dX)) ^ 2 * partial_derive_firstvar_n 2 f (my_proj X)
   - (snd (0,snd dX))^2 * (partial_derive_secondvar_n 2 f (my_proj X))
   - 2*(fst (0, snd dX))*(snd (0,snd dX)) * partial_derive_firstvar (partial_derive_secondvar f) (my_proj X))).
intros; simpl; ring.
apply OuP_implies_OuP2 with 
  (f:=(fun (X:({x| Betw xmin xmax x}*R)) dX => f (add2 (my_proj X) dX) - 2*f (my_proj X) + f (subtract2 (my_proj X) dX) 
           - (fst dX)^2 * (partial_derive_firstvar_n 2 f (my_proj X))
           - (snd dX)^2 * (partial_derive_secondvar_n 2 f (my_proj X))
           - 2*(fst dX)*(snd dX) * partial_derive_firstvar (partial_derive_secondvar f) (my_proj X))) 
  (g:= fun dX => (norm_l2 dX)^4).
intros; auto.
apply DL_discretization_eps_aux with (eps := fun X dX => dX); auto.
exists 1; split; auto with real.
Qed.

End Diff.


