(** R^n *)

Require Export Differential.
Require Export Fourier.
Require Export Reals_compl.
Require ZArith.
Open Local Scope R_scope.

Section Rn.

Variable xmin xmax:R.
Hypothesis xminLtxmax: xmin < xmax.

Definition ni (dx:R): Z :=
  ((jx (xmax-xmin) dx))%Z.


Lemma ni_pos:
   forall dx, 0 < dx -> (0 <= ni dx)%Z.
intros; unfold ni, jx.
apply Int_part_pos.
apply Rle_mult_inv_pos; auto.
apply Rplus_le_reg_l with xmin; ring_simplify.
left; apply xminLtxmax.
Qed.

Lemma ni_Zabs_nat: forall dx, 0 < dx ->
  Z_of_nat (Zabs_nat (ni dx)) = ni dx.
intros.
rewrite inj_Zabs_nat.
apply Zabs_eq.
apply ni_pos; auto.
Qed.


(** R^n is a vector space.
  Vectors in R^n are represented by nat -> R functions *)

Definition add_n (f g : Z -> R) : Z -> R :=
  fun i => f i + g i.

Definition subtract_n (f g : Z -> R) : Z -> R :=
  fun i => f i - g i.

Definition scalar_multiply_n (l : R) (f : Z -> R) : Z -> R :=
  fun i => l * f i.

Notation "f + g" := (add_n f g) : type_scope.
Notation "f - g" := (subtract_n f g) : type_scope.

(** Definition of the dot_product from 0 to n:nat where we add from 0 to n-1 *)


Fixpoint dot_product_n (n:nat) (f g : Z -> R) {struct n} : R :=
  match n with
  | 0 => 0
  | S n' => (f (Z_of_nat n')) * (g (Z_of_nat n')) + dot_product_n n' f g
  end.


Definition squared_norm_n (n:nat) (f : Z -> R) := dot_product_n n f f.

(** Many lemmas *)


Lemma dot_product_n_zero :
  forall u v n,
    (forall i : Z, 
        (0 <= i < Z_of_nat n)%Z -> u i = 0 \/ v i = 0) ->
    dot_product_n n u v = 0.
Proof.
  induction n; auto.
  intros; simpl.
  rewrite IHn.
  case (H (Z_of_nat n)).
 split; auto with zarith.
 rewrite inj_S; auto with zarith.
 intros H1; rewrite H1; ring.
 intros H1; rewrite H1; ring.
 intros i (H1,H2); apply H; rewrite inj_S; split; auto with zarith.
 Qed.


Lemma dot_product_n_distr_l :
  forall (n : nat)  (f g h : Z -> R), 
  dot_product_n n f g + dot_product_n n h g =
  dot_product_n n (add_n f h) g.
intros n; induction n.
intros;unfold add_n;  simpl; ring.
intros; simpl; rewrite <- IHn; unfold add_n; simpl; ring.
Qed.

Lemma dot_product_n_distr_r :
  forall (n : nat) (f g h : Z -> R), 
  dot_product_n n f g + dot_product_n n f h =
  dot_product_n n f (add_n g h).
intros n; induction n.
intros; unfold add_n; simpl; auto with real.
intros; simpl; rewrite <- IHn; unfold add_n; simpl; ring.
Qed.


Lemma dot_product_n_sub_distr_l :
  forall (n : nat) (f g h : Z -> R), 
  dot_product_n n f g - dot_product_n n h g =
  dot_product_n n (subtract_n f h) g.
intros n; induction n.
intros;unfold subtract_n;  simpl; ring.
intros; simpl; rewrite <- IHn; unfold subtract_n; simpl; ring.
Qed.


Lemma dot_product_n_sub_distr_r :
  forall (n : nat)  (f g h : Z -> R), 
  dot_product_n n f g - dot_product_n n f h =
  dot_product_n n f (subtract_n g h).
intros n; induction n.
intros; unfold subtract_n; simpl; ring.
intros; simpl; rewrite <- IHn; unfold subtract_n; simpl; ring.
Qed.


Lemma dot_product_n_scal_mult_l :
  forall (n : nat) (l : R) (f g : Z -> R),
  dot_product_n n (scalar_multiply_n l f) g =
  l * dot_product_n n f g.
intros n; induction n.
intros; unfold scalar_multiply_n; simpl; auto with real.
intros; simpl; rewrite Rmult_plus_distr_l.
rewrite <- IHn; unfold scalar_multiply_n; simpl; ring.
Qed.


Lemma dot_product_n_scal_mult_r :
  forall (n : nat) (l : R) (f g : Z -> R),
  dot_product_n n f (scalar_multiply_n l g) =
  l * dot_product_n n f g.
intros n; induction n.
intros; unfold scalar_multiply_n; simpl; ring.
intros; simpl; rewrite Rmult_plus_distr_l.
rewrite <- IHn; unfold scalar_multiply_n; simpl; ring.
Qed.


Lemma dot_product_n_sym :
  forall (n : nat) (f g : Z -> R),
  dot_product_n n f g = dot_product_n n g f.
intros n; induction n.
simpl; auto with real.
intros; simpl; rewrite <- IHn; simpl; ring.
Qed.


Lemma dot_product_n_eq_suf_l :
  forall (n : nat) (f g h: Z -> R),
  (forall i, (0 <= i < Z_of_nat n)%Z ->  f i = g i) -> 
   dot_product_n n f h = dot_product_n n g h.
intros n; induction n.
intros; simpl; auto with real.
intros; simpl; rewrite IHn with (g:=g); auto with real zarith.
rewrite H; auto with real zarith.
rewrite inj_S; split; auto with zarith.
intros; apply H; auto with zarith.
rewrite inj_S; auto with zarith.
Qed.


Lemma dot_product_n_eq_suf_r :
  forall (n : nat) (f g h: Z -> R),
  (forall i : Z, (0 <= i < Z_of_nat n)%Z -> f i = g i) ->
   dot_product_n n h f = dot_product_n n h g.
intros.
rewrite dot_product_n_sym.
rewrite dot_product_n_eq_suf_l with (1:=H).
apply dot_product_n_sym.
Qed.



Lemma dot_product_n_eq_suf_strict_l :
  forall (n : nat) (f g h: Z -> R),
  (forall i, (0 < i < Z_of_nat n)%Z ->  f i = g i) ->
  h 0%Z = 0 -> 
   dot_product_n n f h = dot_product_n n g h.
intros n; induction n.
intros; simpl; auto with real.
intros; simpl; rewrite IHn with (g:=g); auto with real zarith.
generalize H; case n.
intros; simpl; rewrite H0; ring.
intros n0 H'; rewrite H'; auto with real zarith.
repeat rewrite inj_S; split; auto with zarith.
intros; apply H; auto with zarith.
rewrite inj_S; auto with zarith.
Qed.


Lemma dot_product_n_eq_suf_strict_r :
  forall (n : nat) (f g h: Z -> R),
  (forall i : Z, (0 < i < Z_of_nat n)%Z -> f i = g i) ->
  h 0%Z = 0 -> 
  dot_product_n n h f = dot_product_n n h g.
intros.
rewrite dot_product_n_sym.
rewrite dot_product_n_eq_suf_strict_l with (1:=H) (2:=H0).
apply dot_product_n_sym.
Qed.


Lemma dot_product_n_shift_indices: forall (n:nat) (f1 f2 : Z -> R),
  dot_product_n n f1 f2 = 
        dot_product_n n (fun j => f1 (j+1)%Z) (fun j => f2 (j+1)%Z) 
         + f1 0%Z*f2 0%Z - f1 (Z_of_nat n) * f2 (Z_of_nat n).
intros n; induction n.
intros; simpl; ring.
intros; rewrite inj_S; simpl; rewrite IHn.
unfold Zsucc; ring.
Qed.


Lemma squared_norm_n_pos_aux:
  forall (n:nat) (f: Z -> R),
    0 <= (dot_product_n n f f).
intros n; induction n; simpl; intros.
auto with real.
apply Rle_trans with (0+0); try apply Rplus_le_compat; auto with real.
fold (Rsqr (f (Z_of_nat n))); auto with real.
Qed.


Lemma squared_norm_n_pos:
  forall n (f: Z -> R),
     (0 <= squared_norm_n n f).
intros; unfold squared_norm_n.
apply squared_norm_n_pos_aux.
Qed.


Lemma squared_norm_n_scal_mult_n:  forall n (r:R) (f:Z -> R),
   (squared_norm_n n (scalar_multiply_n r f)) = r*r*(squared_norm_n n f).
intros; unfold squared_norm_n.
rewrite dot_product_n_scal_mult_l; rewrite dot_product_n_scal_mult_r; auto with real.
Qed.

Lemma squared_norm_n_plus:  forall n (f g:Z->R),
  (squared_norm_n n (f+g)%type = squared_norm_n n f + squared_norm_n n g 
  + 2 * dot_product_n n f g)%R.
intros n f g; unfold squared_norm_n.
rewrite <- dot_product_n_distr_l.
repeat rewrite <- dot_product_n_distr_r.
rewrite <- dot_product_n_sym with (f:=f) (g:=g); ring.
Qed.


Lemma squared_norm_n_majoration: forall n f M,
  (forall j,  (0 <= j < Z_of_nat n)%Z -> Rsqr (f j) <= M) ->
  dot_product_n n f f <= INR n  * M.
intros n; induction n.
unfold Rsqr; simpl; intros.
right; ring.
intros; simpl (dot_product_n (S n) f f).
apply Rle_trans with (M + (INR n *M)).
apply Rplus_le_compat.
unfold Rsqr in H; apply H.
rewrite inj_S;split; auto with zarith.
apply IHn.
intros; apply H; rewrite inj_S; split; auto with zarith.
rewrite S_INR; simpl; right; ring. 
Qed.


Lemma Cauchy_Schwarz: forall n f1 f2, 
   Rabs (dot_product_n n f1 f2)
      <= sqrt (squared_norm_n n f1) * sqrt (squared_norm_n n f2).
intros.
assert ((2*dot_product_n n f1 f2) * (2*dot_product_n n f1 f2) - 
  4 * squared_norm_n n f1 * squared_norm_n n f2 <= 0).
apply discr_neg.
intros.
apply Rle_trans with (squared_norm_n n ((scalar_multiply_n x f1)+f2)%type).
apply squared_norm_n_pos.
right.
rewrite squared_norm_n_plus.
rewrite squared_norm_n_scal_mult_n.
rewrite dot_product_n_scal_mult_l.
ring.
apply Rsqr_incr_0_var.
rewrite <- Rsqr_abs.
rewrite Rsqr_mult.
repeat rewrite Rsqr_sqrt; try apply squared_norm_n_pos.
apply Rmult_le_reg_l with (4%R).
fourier.
apply Rplus_le_reg_l with (-4*squared_norm_n n f1 * squared_norm_n n f2).
ring_simplify.
apply Rle_trans with (2:=H).
unfold Rsqr; right; ring.
apply Rmult_le_pos; apply sqrt_pos.
Qed.


Definition finite_dot_product_dx dx u v := dx*dot_product_n (Zabs_nat (ni dx)) u v.

Definition finite_squared_norm_dx dx u :=  finite_dot_product_dx dx u u.

Definition finite_norm_dx dx u :=  sqrt (finite_squared_norm_dx dx u).

(** Lemmas about the  finite dot product *)


Lemma finite_dot_product_dx_zero1:
  forall dx f g, 0 < dx ->
  (forall i, (0 <= i < ni dx)%Z -> f i = 0) -> 
  finite_dot_product_dx dx f g = 0.
intros; unfold finite_dot_product_dx.
rewrite dot_product_n_zero;[apply Rmult_0_r|idtac].
intros i (H1,H2); left; apply H0.
rewrite ni_Zabs_nat in H2; try split; assumption.
Qed.

Lemma finite_dot_product_dx_zero1ou2:
  forall dx f g, 0 < dx ->
  (forall i, (0 <= i < ni dx)%Z -> f i = 0 \/ g i =0) -> 
  finite_dot_product_dx dx f g = 0.
intros; unfold finite_dot_product_dx.
rewrite dot_product_n_zero;[apply Rmult_0_r|idtac].
intros i (H1,H2); apply H0.
rewrite ni_Zabs_nat in H2; try split; assumption.
Qed.


Lemma finite_squared_norm_dx_pos: forall dx f, 
  (0 <= dx) ->
   0 <= finite_squared_norm_dx dx f.
intros.
unfold finite_squared_norm_dx, finite_dot_product_dx.
apply Rmult_le_pos; auto.
apply squared_norm_n_pos_aux.
Qed.

Lemma finite_squared_norm_dx_eq: forall dx f1 f2,  0 < dx ->
  (forall i, (0 <= i < ni dx)%Z ->  f1 i = f2 i) -> 
  finite_squared_norm_dx dx f1 = finite_squared_norm_dx dx f2.
intros; unfold finite_squared_norm_dx, finite_dot_product_dx.
rewrite <- (ni_Zabs_nat _ H) in H0.
rewrite dot_product_n_eq_suf_l with (1:=H0).
rewrite dot_product_n_eq_suf_r with (1:=H0); reflexivity.
Qed.

Lemma finite_squared_norm_dx_zero: forall dx f, 0 < dx ->
   (forall i, (0 <= i < ni dx)%Z -> f i = 0) -> 
  finite_squared_norm_dx dx f = 0.
intros; unfold finite_squared_norm_dx.
rewrite finite_dot_product_dx_zero1; auto.
Qed.

Lemma finite_squared_norm_dx_homogen: forall dx l f, 
  finite_squared_norm_dx dx (scalar_multiply_n l f) = 
    l * l * finite_squared_norm_dx dx f.
intros; unfold finite_squared_norm_dx,finite_dot_product_dx.
rewrite dot_product_n_scal_mult_r.
rewrite dot_product_n_scal_mult_l.
ring.
Qed.


Lemma finite_Cauchy_Schwarz_dx: forall dx f1 f2,
   0 <= dx ->
   Rabs (finite_dot_product_dx dx f1 f2)
      <= finite_norm_dx dx f1 * finite_norm_dx dx f2.
intros dx f1 f2 hdx.
unfold finite_norm_dx, finite_squared_norm_dx, finite_dot_product_dx.
repeat rewrite  sqrt_mult; trivial; try apply squared_norm_n_pos.
apply Rle_trans with ((sqrt dx * sqrt dx)* 
  (sqrt (dot_product_n (Zabs_nat (ni dx)) f1 f1) * 
  sqrt (dot_product_n (Zabs_nat (ni dx)) f2 f2))).
2: right; ring.
rewrite Rabs_mult.
apply Rmult_le_compat; try apply Rabs_pos.
rewrite <- sqrt_mult; replace (dx*dx) with dx²;auto with real.
right; apply sym_eq; apply sqrt_Rsqr_abs.
apply Cauchy_Schwarz.
Qed.

Lemma finite_squared_norm_dx_triang: forall dx f1 f2, 
  0 <= dx -> 
  finite_norm_dx dx (fun j => f1 j + f2 j) <= 
      finite_norm_dx dx f1 + finite_norm_dx dx f2.
intros dx f1 f2 hdx.
unfold finite_norm_dx, finite_squared_norm_dx, finite_dot_product_dx.
apply Rsqr_incr_0_var.
rewrite Rsqr_sqrt.
unfold Rsqr; rewrite Rmult_plus_distr_l.
rewrite Rmult_plus_distr_r; rewrite Rmult_plus_distr_r.
repeat rewrite sqrt_sqrt.
change (dx * squared_norm_n (Zabs_nat (ni dx)) (add_n f1 f2)
  <= dx * squared_norm_n (Zabs_nat (ni dx)) f1 +
      sqrt (dx * squared_norm_n (Zabs_nat (ni dx)) f2) *
     sqrt (dx * squared_norm_n (Zabs_nat (ni dx))f1) +
     (sqrt (dx * squared_norm_n (Zabs_nat (ni dx))f1) *
     sqrt (dx * squared_norm_n (Zabs_nat (ni dx))f2) +
  dx * squared_norm_n (Zabs_nat (ni dx)) f2)).
rewrite squared_norm_n_plus.
apply Rplus_le_reg_l with (-dx*squared_norm_n (Zabs_nat (ni dx)) f2
   - dx*squared_norm_n (Zabs_nat (ni dx)) f1).
ring_simplify.
repeat rewrite sqrt_mult; trivial.
repeat rewrite Rmult_assoc.
apply Rmult_le_compat_l.
auto with real.
replace (Rmult (sqrt dx)
     (Rmult (sqrt (squared_norm_n (Zabs_nat (ni dx)) f2))
        (Rmult (sqrt dx) (sqrt (squared_norm_n (Zabs_nat (ni dx)) f1)))))
  with
  ((sqrt dx * sqrt dx) * ((sqrt (squared_norm_n (Zabs_nat (ni dx)) f2)) *
     sqrt (squared_norm_n (Zabs_nat (ni dx)) f1))) by ring.
rewrite sqrt_sqrt; trivial.
apply Rmult_le_compat_l; trivial.
apply Rle_trans with (1:= RRle_abs (dot_product_n (Zabs_nat (ni dx)) f1 f2)).
rewrite Rmult_comm.
apply Cauchy_Schwarz.
apply squared_norm_n_pos.
apply squared_norm_n_pos.
apply Rmult_le_pos; trivial. apply squared_norm_n_pos.
apply Rmult_le_pos; trivial. apply squared_norm_n_pos.
apply Rmult_le_pos; trivial. apply squared_norm_n_pos.
apply Rle_trans with (0+0); auto with real; apply Rplus_le_compat.
apply sqrt_pos.
apply sqrt_pos.
Qed.


Lemma finite_squared_norm_dx_majoration: forall dx f M,
   0 < dx -> 0 <= M ->
   (forall j : Z, (0 <= j < ni dx )%Z -> Rsqr (f j) <= M) ->
   finite_squared_norm_dx dx f <= (xmax-xmin) * M.
Proof.
intros.
unfold finite_squared_norm_dx, finite_dot_product_dx.
apply Rle_trans with (dx*(INR (Zabs_nat (ni dx))*M))%R.
apply Rmult_le_compat_l; auto with real.
apply squared_norm_n_majoration.
rewrite ni_Zabs_nat; auto.
rewrite INR_IZR_INZ; rewrite ni_Zabs_nat; auto.
rewrite <- Rmult_assoc; apply Rmult_le_compat_r; trivial.
unfold ni, jx.
apply Rle_trans with (dx * ((((xmax - xmin) / dx)))).
apply Rmult_le_compat_l; auto with real.
apply (proj1 (base_Int_part _)).
right; field; auto with real.
Qed.


Lemma finite_dot_product_dx_sym: forall dx u v, 
   finite_dot_product_dx dx u v = finite_dot_product_dx dx v u.
intros; unfold finite_dot_product_dx.
apply f_equal.
apply dot_product_n_sym.
Qed.

Lemma finite_dot_product_dx_eq_l: forall dx u v w, 0 < dx ->
   (forall i, (0 <= i < ni dx)%Z -> u i = v i) ->
   finite_dot_product_dx dx u w = finite_dot_product_dx dx v w.
intros; unfold finite_dot_product_dx.
apply f_equal.
apply dot_product_n_eq_suf_l.
rewrite ni_Zabs_nat; auto.
Qed.

Lemma finite_dot_product_dx_eq_r: forall dx u v w, 0 < dx ->
   (forall i, (0 <= i < ni dx)%Z -> u i = v i) ->
   finite_dot_product_dx dx w u = finite_dot_product_dx dx w v.
intros; unfold finite_dot_product_dx.
apply f_equal.
apply dot_product_n_eq_suf_r.
rewrite ni_Zabs_nat; auto.
Qed.

Lemma finite_dot_product_dx_eq_strict_l: forall dx u v w, 0 < dx ->
   (forall i, (0 < i < ni dx)%Z -> u i = v i) ->
   w 0%Z = 0 ->
   finite_dot_product_dx dx u w = finite_dot_product_dx dx v w.
intros; unfold finite_dot_product_dx.
apply f_equal.
apply dot_product_n_eq_suf_strict_l; auto.
rewrite ni_Zabs_nat; auto.
Qed.

Lemma finite_dot_product_dx_eq_strict_r: forall dx u v w, 0 < dx ->
   (forall i, (0 < i < ni dx)%Z -> u i = v i) ->
   w 0%Z = 0 ->
   finite_dot_product_dx dx w u = finite_dot_product_dx dx w v.
intros; unfold finite_dot_product_dx.
apply f_equal.
apply dot_product_n_eq_suf_strict_r; auto.
rewrite ni_Zabs_nat; auto.
Qed.

Lemma finite_dot_product_dx_add_l: forall dx u v w,
   finite_dot_product_dx dx (add_n u v) w=
   finite_dot_product_dx dx u w +  finite_dot_product_dx dx v w.
intros; unfold finite_dot_product_dx.
rewrite <- Rmult_plus_distr_l.
apply f_equal.
apply sym_eq; apply dot_product_n_distr_l.
Qed.


Lemma finite_dot_product_dx_add_r: forall dx u v w,
   finite_dot_product_dx dx w (add_n u v) =
   finite_dot_product_dx dx w u +  finite_dot_product_dx dx w v.
intros; unfold finite_dot_product_dx.
rewrite <- Rmult_plus_distr_l.
apply f_equal.
apply sym_eq; apply dot_product_n_distr_r.
Qed.

Lemma finite_dot_product_dx_sub_l: forall dx u v w,
   finite_dot_product_dx dx (subtract_n u v) w =
   finite_dot_product_dx dx u w - finite_dot_product_dx dx v w.
intros; unfold finite_dot_product_dx.
rewrite <- Rmult_minus_distr_l.
apply f_equal.
apply sym_eq; apply dot_product_n_sub_distr_l.
Qed.

Lemma finite_dot_product_dx_sub_r: forall dx u v w, 
   finite_dot_product_dx dx w (subtract_n u v) =
   finite_dot_product_dx dx w u - finite_dot_product_dx dx w v.
intros; unfold finite_dot_product_dx.
rewrite <- Rmult_minus_distr_l.
apply f_equal.
apply sym_eq; apply dot_product_n_sub_distr_r.
Qed.


Lemma finite_dot_product_dx_scal_mult_l :
  forall dx u v l, 
  finite_dot_product_dx dx (scalar_multiply_n l u) v =
  l * finite_dot_product_dx dx u v.
intros; unfold finite_dot_product_dx.
rewrite dot_product_n_scal_mult_l.
ring.
Qed.

Lemma finite_dot_product_dx_scal_mult_r :
  forall dx u v l,  
  finite_dot_product_dx dx u (scalar_multiply_n l v) =
  l * finite_dot_product_dx dx u v.
intros; unfold finite_dot_product_dx.
rewrite dot_product_n_scal_mult_r.
ring.
Qed.


Lemma finite_dot_product_dx_shift_indices: forall dx (f1 f2 : Z -> R),
  0 < dx ->
  f1 0%Z = 0 \/ f2 0%Z = 0 ->
  f1 (ni dx) = 0 \/ f2 (ni dx) = 0 ->
  finite_dot_product_dx dx f1 f2 = 
        finite_dot_product_dx dx (fun j => f1 (j+1)%Z) (fun j => f2 (j+1)%Z).
intros; unfold finite_dot_product_dx.
rewrite dot_product_n_shift_indices.
rewrite ni_Zabs_nat; trivial.
destruct H0; destruct H1; rewrite H0, H1; ring.
Qed.



(* Lemme sur Ah et finite_dot_product_dx *)

Lemma semi_finite_dot_product_dx: forall dx Ah u v,
  (forall u v, 0 < dx -> u 0%Z = 0 -> u (ni dx) = 0 -> v 0%Z = 0 -> v (ni dx) = 0 ->
    finite_dot_product_dx dx (Ah u) v
   = finite_dot_product_dx dx u (Ah v)) ->
  0 < dx -> u 0%Z = 0 -> u (ni dx) = 0 -> v 0%Z = 0 -> v (ni dx) = 0 ->
  finite_dot_product_dx dx (Ah u) v =
    /4 * finite_dot_product_dx dx (Ah (add_n u v)) (add_n u v)
   - /4 * finite_dot_product_dx dx (Ah (subtract_n u v)) (subtract_n u v).
intros dx Ah u v H K; intros.
rewrite finite_dot_product_dx_sub_r.
rewrite finite_dot_product_dx_add_r.
repeat rewrite H; auto.
repeat rewrite finite_dot_product_dx_sub_l.
repeat rewrite finite_dot_product_dx_add_l.
repeat rewrite <- (H u); auto.
repeat rewrite <- finite_dot_product_dx_sym with (v:= Ah u).
field.
unfold subtract_n; rewrite H0,H2; ring.
unfold subtract_n; rewrite H1,H3; ring.
unfold subtract_n; rewrite H0,H2; ring.
unfold subtract_n; rewrite H1,H3; ring.
unfold add_n; rewrite H0,H2; ring.
unfold add_n; rewrite H1,H3; ring.
unfold add_n; rewrite H0,H2; ring.
unfold add_n; rewrite H1,H3; ring.
Qed.


Lemma finite_squared_norm_parall: forall dx x y, 
 finite_squared_norm_dx dx (x+y)%type +
  finite_squared_norm_dx dx (x-y)%type =
   2*(finite_squared_norm_dx dx x + finite_squared_norm_dx dx y).
intros; unfold finite_squared_norm_dx.
rewrite finite_dot_product_dx_add_r.
repeat rewrite finite_dot_product_dx_add_l.
rewrite finite_dot_product_dx_sub_r.
repeat rewrite finite_dot_product_dx_sub_l.
ring.
Qed.

Lemma finite_squared_norm_parall_add: forall dx x y, 
   0 <= dx ->
   finite_squared_norm_dx dx (x+y)%type <= 
  2*(finite_squared_norm_dx dx x + finite_squared_norm_dx dx y).
intros.
rewrite <- finite_squared_norm_parall.
rewrite <- (Rplus_0_r (finite_squared_norm_dx dx (x + y)%type)) at 1.
apply Rplus_le_compat_l.
apply finite_squared_norm_dx_pos; auto.
Qed.

Lemma finite_squared_norm_parall_sub: forall dx x y, 
   0 <= dx -> 
   finite_squared_norm_dx dx (x-y)%type <= 
  2*(finite_squared_norm_dx dx x + finite_squared_norm_dx dx y).
intros.
rewrite <- finite_squared_norm_parall.
rewrite <- (Rplus_0_l (finite_squared_norm_dx dx (x - y)%type)) at 1.
apply Rplus_le_compat_r.
apply finite_squared_norm_dx_pos; auto.
Qed.


Lemma dot_product_n_le: forall n uh vh r1 r2,
  (forall i, (0 <= i < Z_of_nat n)%Z -> Rabs (uh i) <= r1) ->
  (forall i, (0 <= i < Z_of_nat n)%Z -> Rabs (vh i) <= r2) ->
  Rabs (dot_product_n n uh vh) <= INR n*r1*r2.
intros.
induction n.
simpl; rewrite Rabs_R0; right; ring.
rewrite inj_S in H, H0.
simpl (dot_product_n (S n) uh vh).
apply Rle_trans with (1:=Rabs_triang _ _).
rewrite S_INR, 2!Rmult_plus_distr_r, Rplus_comm.
apply Rplus_le_compat.
apply IHn; auto with zarith.
rewrite Rmult_1_l, Rabs_mult.
apply Rmult_le_compat; try apply Rabs_pos.
apply H; auto with zarith.
apply H0; auto with zarith.
Qed.

Lemma dot_product_n_le2: forall n uh vh r1 r2,
  0 <= r1 -> 0 <= r2 ->
  (forall i, (0 <= i < Z_of_nat n)%Z -> Rabs (uh i) <= r1 \/ vh i=0) ->
  (forall i, (0 <= i < Z_of_nat n)%Z -> Rabs (vh i) <= r2 \/ uh i=0) ->
  Rabs (dot_product_n n uh vh) <= INR n*r1*r2.
intros.
induction n.
simpl; rewrite Rabs_R0; right; ring.
rewrite inj_S in H1, H2.
simpl (dot_product_n (S n) uh vh).
apply Rle_trans with (1:=Rabs_triang _ _).
rewrite S_INR, 2!Rmult_plus_distr_r, Rplus_comm.
apply Rplus_le_compat.
apply IHn; auto with zarith.
rewrite Rmult_1_l, Rabs_mult.
case (H1 (Z_of_nat n)); auto with zarith.
case (H2 (Z_of_nat n)); auto with zarith.
intros; apply Rmult_le_compat; try apply Rabs_pos; auto.
intros Y1 Y2; rewrite Y1, Rabs_R0, Rmult_0_l.
apply Rmult_le_pos; auto.
intros Y1; rewrite Y1, Rabs_R0, Rmult_0_r.
apply Rmult_le_pos; auto.
Qed.

Lemma finite_dot_product_le: forall dx uh vh r1 r2, 0 < dx ->
  ((0 < ni dx)%Z \/ (0 <= r1 /\ 0 <= r2)) ->
  (forall i, (0 <= i < ni dx )%Z -> Rabs (uh i) <= r1) ->
  (forall i, (0 <= i < ni dx )%Z -> Rabs (vh i) <= r2) ->
  Rabs (finite_dot_product_dx dx uh vh) <= (xmax-xmin)*r1*r2.
intros.
unfold finite_dot_product_dx.
rewrite Rabs_mult.
rewrite Rabs_right; try apply Rle_ge; auto with real.
apply Rle_trans with (dx*(INR (Zabs_nat (ni dx))*r1*r2)).
apply Rmult_le_compat_l; auto with real.
apply dot_product_n_le.
rewrite ni_Zabs_nat; auto.
rewrite ni_Zabs_nat; auto.
rewrite INR_IZR_INZ, ni_Zabs_nat; auto.
rewrite <- Rmult_assoc.
apply Rmult_le_compat_r.
case H0; intros.
apply Rle_trans with (Rabs (vh 0%Z)); auto with zarith real; apply Rabs_pos.
apply H3.
rewrite <- Rmult_assoc.
apply Rmult_le_compat_r.
case H0; intros.
apply Rle_trans with (Rabs (uh 0%Z)); auto with zarith real; apply Rabs_pos.
apply H3.
apply Rmult_le_reg_l with (/dx); auto with real.
apply Rle_trans with (IZR (ni dx)).
right; field; auto with real.
rewrite Rmult_comm.
unfold ni, jx.
apply (proj1 (base_Int_part _)).
Qed.

Lemma finite_dot_product_le2: forall dx uh vh r1 r2, 0 < dx ->
  0 <= r1 ->  0 <= r2 ->
  (forall i, (0 <= i < ni dx )%Z -> Rabs (uh i) <= r1 \/ vh i=0) ->
  (forall i, (0 <= i < ni dx )%Z -> Rabs (vh i) <= r2 \/ uh i=0) ->
  Rabs (finite_dot_product_dx dx uh vh) <= (xmax-xmin)*r1*r2.
intros.
unfold finite_dot_product_dx.
rewrite Rabs_mult.
rewrite Rabs_right; try apply Rle_ge; auto with real.
apply Rle_trans with (dx*(INR (Zabs_nat (ni dx))*r1*r2)).
apply Rmult_le_compat_l; auto with real.
apply dot_product_n_le2; auto.
rewrite ni_Zabs_nat; auto.
rewrite ni_Zabs_nat; auto.
rewrite INR_IZR_INZ, ni_Zabs_nat; auto.
rewrite <- Rmult_assoc.
apply Rmult_le_compat_r; auto.
rewrite <- Rmult_assoc.
apply Rmult_le_compat_r; auto.
apply Rmult_le_reg_l with (/dx); auto with real.
apply Rle_trans with (IZR (ni dx)).
right; field; auto with real.
rewrite Rmult_comm.
unfold ni, jx.
apply (proj1 (base_Int_part _)).
Qed.



Lemma finite_norm_le: forall dx uh r, 0 < dx ->  ((0 < ni dx)%Z \/ (0 <= r)) ->
  (forall i, (0 <= i < ni dx )%Z -> Rabs (uh i) <= r) ->
     finite_norm_dx dx uh <= sqrt (xmax -xmin)*r.
intros.
unfold finite_norm_dx.
apply Rsqr_incr_0_var.
rewrite Rsqr_sqrt.
unfold finite_squared_norm_dx.
apply Rle_trans with (1:=RRle_abs _).
apply Rle_trans with ((xmax-xmin)*r*r).
apply finite_dot_product_le; auto with real zarith.
case H0; auto.
rewrite Rsqr_mult.
rewrite Rsqr_sqrt.
right; unfold Rsqr; ring.
apply Rplus_le_reg_l with xmin; ring_simplify; auto with real.
apply finite_squared_norm_dx_pos; auto with real.
apply Rmult_le_pos.
apply sqrt_pos.
case H0; auto; intros.
apply Rle_trans with (Rabs (uh 0%Z)); auto with zarith real; apply Rabs_pos.
Qed.


Lemma dot_product_n_le_reverse:  forall n uh i,  (0 <= i < Z_of_nat n)%Z ->
   Rsqr (uh i) <= dot_product_n n uh uh.
intros; induction n.
simpl; simpl in H.
absurd (0 < 0)%Z; auto with zarith.
simpl.
case (Zle_lt_or_eq i (Z_of_nat n)).
rewrite inj_S in H; auto with zarith.
intros H2; apply Rle_trans with (dot_product_n n uh uh).
apply IHn; auto with zarith.
rewrite <- (Rplus_0_l (dot_product_n n uh uh)) at 1.
apply Rplus_le_compat_r.
fold (Rsqr (uh (Z_of_nat n))); auto with real.
intros H2; rewrite <- H2.
rewrite <- (Rplus_0_r (Rsqr (uh i))).
unfold Rsqr; apply Rplus_le_compat_l.
apply squared_norm_n_pos_aux.
Qed.


Lemma finite_norm_le_reverse: forall dx uh r,  0 < dx ->
    finite_norm_dx dx uh <= r ->  forall i, (0 <= i < ni dx)%Z -> Rabs (uh i) <= r / sqrt (dx).
intros.
assert (0 <= r).
apply Rle_trans with (2:=H0).
unfold finite_norm_dx; apply sqrt_pos.
apply Rmult_le_reg_l with (sqrt dx).
now apply sqrt_lt_R0.
apply Rle_trans with r.
2: right; field.
2: specialize (sqrt_lt_R0 _ H); auto with real.
apply Rle_trans with (2:=H0).
unfold finite_norm_dx, finite_squared_norm_dx, finite_dot_product_dx.
apply Rsqr_incr_0_var.
2: apply sqrt_pos.
rewrite Rsqr_sqrt.
rewrite Rsqr_mult.
rewrite Rsqr_sqrt; auto with real.
apply Rmult_le_compat_l; auto with real.
rewrite <- Rsqr_abs.
apply dot_product_n_le_reverse.
rewrite ni_Zabs_nat; auto.
apply Rmult_le_pos; auto with real.
apply squared_norm_n_pos_aux.
Qed.


End Rn.
