(** Some needed definitions and lemmas about real numbers *)

Require Export Reals.
Open Local Scope nat_scope.
Open Local Scope R_scope.

(** Discretization of real values to integers or to natural numbers *)

Definition jx (x:R) (dx:R) : Z :=
   (Int_part (x/dx)).

Definition kt (t:R) (dt:R) : nat :=
   Zabs_nat (Int_part (t/dt)).

(** Missing lemmas *)

Lemma Rmult_eq_compat_r : forall r r1 r2, r1 = r2 -> r1 * r = r2 * r.
intros; rewrite H; ring.
Qed.

Lemma Rplus_eq_compat_r
     : forall r r1 r2 : R, r1 = r2 -> r1 + r = r2 + r.
intros; now rewrite H. 
Qed.


Lemma sqrt_pos: forall x, 0 <= sqrt x.
intros; unfold sqrt.
case (Rcase_abs x).
intros; now right.
intros.
apply Rsqrt_positivity.
Qed.


(** About Int_part *)

Lemma Int_part_pos: forall r, 0 <= r -> (0 <= (Int_part r))%Z.
intros.
assert (-1 < Int_part r)%Z; auto with zarith.
apply lt_IZR; simpl.
apply Rlt_le_trans with (IZR (Int_part r) - r)%R.
eapply base_Int_part.
apply Rle_trans with (IZR (Int_part r) - 0)%R; [idtac|right; ring].
unfold Rminus; apply Rplus_le_compat_l; auto with real.
Qed.

Lemma Int_part_IZR: forall n : Z, Int_part (IZR n) = n.
intros n; elim (base_Int_part (IZR n)); intros Y1 Y2.
assert (Int_part (IZR n) <= n)%Z by now apply le_IZR.
assert (n -1 < Int_part (IZR n))%Z; auto with zarith.
apply lt_IZR.
unfold Zminus; rewrite plus_IZR; simpl.
apply Rplus_lt_reg_r with (-IZR n).
apply Rle_lt_trans with (-1).
right; ring.
apply Rlt_le_trans with (IZR (Int_part (IZR n)) - IZR n).
auto with real.
right; ring.
Qed.

Lemma Int_part_monotone: forall r1 r2,
  r1 <= r2 ->  (Int_part r1 <= Int_part r2)%Z.
intros.
case (Zle_or_lt (Int_part r1) (Int_part r2)); trivial.
intros H1; absurd (r1 <= r2); auto.
apply Rlt_not_le.
apply Rlt_le_trans with (IZR (Int_part r2) + 1).
apply Rplus_lt_reg_r with (-r2-1)%R; ring_simplify.
apply Rlt_le_trans with (IZR (Int_part r2) - r2).
eapply base_Int_part.
right; ring.
apply Rle_trans with (IZR (Int_part r1)).
apply Rle_trans with (IZR (Int_part r2+1)).
rewrite plus_IZR; auto with real.
apply IZR_le.
auto with zarith.
eapply base_Int_part.
Qed.

Lemma Rminus_Int_part_le:
  forall r1 r2 : R,
  (Int_part (r1 - r2) <= Int_part r1 - Int_part r2)%Z.
intros.
case (Rle_or_lt (frac_part r2) (frac_part r1)); intros.
rewrite Rminus_Int_part1; auto with real zarith.
rewrite Rminus_Int_part2; auto with real zarith.
Qed.

Lemma Rplus_Int_part_ge:
  forall r1 r2 : R,
  (Int_part r1 + Int_part r2 <= Int_part (r1 + r2))%Z.
intros.
case (Rle_or_lt 1 (frac_part r1+frac_part r2)); intros.
rewrite plus_Int_part1; auto with real zarith.
rewrite plus_Int_part2; auto with real zarith.
Qed.

Lemma up_gt_zero: forall r,
  0 < r -> (0 < up r)%Z.
intros.
apply lt_IZR.
apply Rlt_trans with (1:=H).
eapply archimed.
Qed.

(** About discretizations *)

Lemma kt_eq: forall dt i, dt<>0 -> kt (INR i*dt) dt = i.
intros; unfold kt.
replace (INR i * dt / dt) with (INR i) by (field; trivial).
rewrite INR_IZR_INZ.
rewrite Int_part_IZR.
now rewrite Zabs_nat_Z_of_nat.
Qed.

Lemma kt_le: forall t dt, 0 <= t -> 0 < dt -> INR(kt t dt) <= t/dt.
intros; unfold kt.
destruct (base_Int_part (t / dt)).
apply Rle_trans with (2:=H1).
rewrite INR_IZR_INZ.
rewrite inj_Zabs_nat.
rewrite Zabs_eq.
now right.
apply Int_part_pos.
unfold Rdiv; apply Rmult_le_pos; auto with real.
Qed.

(** Finite sum from s to n, whose value is zero if s > n *)

Definition sum_f (s n:nat) (f: nat -> R) :=
   if (le_gt_dec s n) 
      then sum_f_R0 (fun x : nat => f (x + s)%nat) (n - s)
      else 0.

Lemma sum_f_triang : forall s n f,
  Rabs (sum_f s n f) <= sum_f s n (fun x => Rabs (f x)).
intros.
unfold sum_f; case (le_gt_dec s n); intros.
apply Rsum_abs.
rewrite Rabs_R0.
now apply Rle_refl.
Qed.

Lemma sum_f_pos: forall s n f,
  (forall i, 0 <= f i)
   -> 0 <= sum_f s n (fun x => f x).
intros.
unfold sum_f; case (le_gt_dec s n); intros.
apply cond_pos_sum.
intros; apply H.
auto with real.
Qed.


Lemma sum_f_le: forall s n f M,
  (forall i, (s <= i <= n)%nat -> Rabs (f i) <= M)
   -> 0 <= M
   -> sum_f s n (fun x => Rabs (f x)) <= (INR (n-s)+1)*M.
intros.
unfold sum_f; case (le_gt_dec s n); intros.
assert (s <= (n-s) +s <= n)%nat by omega.
induction (n-s)%nat.
simpl.
ring_simplify ((0+1)*M).
apply H; auto.
rewrite S_INR.
simpl.
apply Rle_trans with ((INR n0 + 1) * M + M).
apply Rplus_le_compat.
apply IHn0.
omega.
now apply H.
right; ring.
rewrite <- S_INR.
apply Rmult_le_pos; auto with arith real.
Qed.

Lemma sum_f_Rabs_minus1: forall s n f, 
   Rabs (sum_f s (n-1) f) <= Rabs (sum_f s n f) + Rabs (f n).
intros s n; case n.
simpl.
unfold sum_f; case (le_gt_dec s 0).
simpl; intros.
apply Rle_trans with (Rabs (f s) + 0).
right; ring.
apply Rplus_le_compat_l.
apply Rabs_pos.
intros; rewrite Rabs_R0; ring_simplify.
apply Rabs_pos.
intros; unfold sum_f.
replace (S n0-1)%nat with n0 by omega.
case (le_gt_dec s n0); intros.
case (le_gt_dec s (S n0)); intros.
rewrite sum_N_predN with (N:=(S n0-s)%nat).
replace (pred (S n0 - s)) with (n0-s)%nat by omega.
replace (S n0-s+s)%nat with (S n0) by omega.
apply Rle_trans with ((Rabs (sum_f_R0 (fun x : nat => f (x + s)%nat) (n0- s)) -Rabs (f (S n0))) + Rabs (f (S n0))).
right; ring.
apply Rplus_le_compat_r.
rewrite <- (Rabs_Ropp (f (S n0))).
apply Rle_trans with (1:=Rabs_triang_inv _ _).
right; apply f_equal; ring.
omega.
absurd (s > S n0)%nat; trivial; omega.
rewrite Rabs_R0.
replace 0 with (0+0) by ring.
apply Rplus_le_compat; apply Rabs_pos.
Qed.

Lemma sum_f_S: forall s n f, 
  (s <= S n)%nat ->
  sum_f s (S n) f = sum_f s n f + f (S n).
intros; unfold sum_f.
case (le_gt_dec s (S n)).
intros _.
case (le_gt_dec s n).
intros.
replace (S n - s)%nat with (S (n-s)) by omega.
simpl.
now replace (n-s+s)%nat with n by omega.
intros.
replace (S n-s)%nat with 0%nat by omega.
simpl.
replace (S n) with s by omega.
ring.
intros; absurd (s <= S n)%nat; omega.
Qed.

Lemma sum_f_S_Rabs_minus1: forall s n f, 
   Rabs (sum_f s (n-1) (fun j => f (S j))) <= Rabs (sum_f s n f) + Rabs (f s) + Rabs (f 1%nat).
intros s n; case n.
simpl.
unfold sum_f; case (le_gt_dec s 0).
simpl; intros.
replace s with 0%nat by omega.
apply Rle_trans with (0+0+Rabs (f 1%nat)).
right; ring.
apply Rplus_le_compat_r.
apply Rplus_le_compat;apply Rabs_pos.
intros; rewrite Rabs_R0; ring_simplify.
replace 0 with (0+0) by ring.
apply Rplus_le_compat;apply Rabs_pos.
intros; unfold sum_f.
replace (S n0-1)%nat with n0 by omega.
case (le_gt_dec s n0); intros.
case (le_gt_dec s (S n0)); intros.
rewrite decomp_sum with (N:=(S n0 - s)%nat).
replace (pred (S n0 - s)) with (n0-s)%nat by omega.
replace (0+s)%nat with s by omega.
apply Rle_trans with 
  ((Rabs (sum_f_R0 (fun x : nat => f (S x + s)%nat) (n0- s)) -Rabs (f s)) + Rabs (f s)).
right; ring_simplify.
apply f_equal.
apply sum_eq.
intros; apply f_equal; omega.
apply Rle_trans with 
 (Rabs (f s + sum_f_R0 (fun i : nat => f (S i + s)%nat) (n0 - s)) + Rabs (f s)).
apply Rplus_le_compat_r.
rewrite <- (Rabs_Ropp (f s)).
apply Rle_trans with (1:=Rabs_triang_inv _ _).
right; apply f_equal; ring.
rewrite Rplus_assoc; apply Rplus_le_compat_l.
apply Rle_trans with (Rabs (f s)+0).
right; ring.
apply Rplus_le_compat_l; apply Rabs_pos.
omega.
absurd (s > S n0)%nat; trivial; omega.
rewrite Rabs_R0.
replace 0 with (0+0+0) by ring.
repeat apply Rplus_le_compat; apply Rabs_pos.
Qed.


(** If a degree-2 polynomial is always nonnegative, its dicriminant is nonpositive *)

Lemma discr_neg: forall a b c,
  (forall x, 0 <= a*x*x+b*x+c) ->
     b*b-4*a*c <= 0.
intros.
case (Req_dec a 0); intros H'.
cut (b=0).
intros H1; rewrite H'; rewrite H1.
right; ring.
case (Req_dec b 0); trivial; intros.
absurd (0 <= a*((-c-1)/b)*((-c-1)/b)+b*((-c-1)/b)+c).
2: apply H.
apply Rlt_not_le.
rewrite H'; apply Rle_lt_trans with (-1).
right; field; trivial.
apply Ropp_lt_gt_0_contravar; apply Rlt_gt; auto with real.
assert (0 <= c).
apply Rle_trans with (a*0*0+b*0+c);[idtac|right; ring].
apply H.
assert (0 < a).
cut (0 <= a).
intros T; case T;auto with real.
intros T'; absurd (a=0); auto.
case (Req_dec b 0); intros.
case (Rle_or_lt 0 a); trivial; intros.
absurd (0 <= a* sqrt ((c+1)/(-a)) * sqrt ((c+1)/(-a)) +b*sqrt ((c+1)/(-a))+c).
2: apply H.
apply Rlt_not_le.
rewrite H1; ring_simplify ( 0 * sqrt ((c + 1) / - a)).
rewrite Rmult_assoc.
rewrite sqrt_sqrt.
apply Rle_lt_trans with (-1).
right; field; auto with real.
apply Ropp_lt_gt_0_contravar; apply Rlt_gt; auto with real.
unfold Rdiv; apply Rmult_le_pos; auto with real.
case H0; intros.
apply Rmult_le_reg_l with (c*c/(b*b)).
unfold Rdiv; apply Rmult_lt_0_compat.
apply Rmult_lt_0_compat; trivial.
apply Rinv_0_lt_compat.
fold (Rsqr b); auto with real.
ring_simplify.
apply Rle_trans with (a*(-c/b)*(-c/b)+b*(-c/b)+c).
apply H.
right; field; trivial.
apply Rmult_le_reg_l with (b*b).
fold (Rsqr b); auto with real.
ring_simplify.
apply Rle_trans with (Rsqr b); auto with real.
apply Rplus_le_reg_l with (-Rsqr b); ring_simplify.
apply Rle_trans with (a*(-b)*(-b)+b*(-b)+c).
apply H.
rewrite <- H2; unfold Rsqr; right; ring.
case (Rle_or_lt (b * b - 4 * a * c) 0); trivial.
intros H2.
absurd ( 0 <= a * (-b/(2*a)) * (-b/(2*a)) + b * (-b/(2*a)) + c).
apply Rlt_not_le.
replace (a * (- b / (2*a)) * (- b / (2*a)) + b * (- b / (2*a)) + c) with
  (-b*b/(4*a)+c).
apply Rmult_lt_reg_l with (4*a).
repeat apply Rmult_lt_0_compat; auto with real.
apply Rplus_lt_reg_r with (b*b-4*a*c).
apply Rle_lt_trans with 0%R.
right; field; auto.
apply Rlt_le_trans with (1:=H2); right; ring.
field; auto.
apply H.
Qed.



