(** The wave equation: its definition, its discretization *)

Require Export Differential.
Require Export R_n.
Require Export ZArith.
Open Local Scope nat_scope.
Open Local Scope R_scope.


Section Wave.
Variable xmin xmax T:R.

Hypothesis xminLtxmax: xmin < xmax.
Hypothesis T_pos: 0 < T.

Variable partial_derive_firstvar: (R * R -> R) -> (R * R -> R).
Variable partial_derive_secondvar: (R * R -> R) -> (R * R -> R).

Definition Xjk (dX : R * R) (j: Z) (k:nat) : R * R :=
  (xmin+(IZR j)*(fst dX), (INR k)*(snd dX)).

Definition Xxt (X dX:R*R) : R*R :=
  Xjk dX (jx (fst X-xmin) (fst dX)) (kt (snd X) (snd dX)).

Notation ni x := (ni xmin xmax x).
Notation finite_dot_product_dx := (fun dx u v => (finite_dot_product_dx xmin xmax dx u v)).
Notation finite_squared_norm_dx := (fun dx u => (finite_squared_norm_dx xmin xmax dx u)).
Notation finite_norm_dx := (fun dx u => (finite_norm_dx xmin xmax dx u)).

(** 1D wave equation *)

Variable c:R.
Variable u0 u1: R->R.
Variable f : R*R -> R.
Variable usol : R * R -> R.

Variable u0h u1h: R*R -> Z-> R.
Variable fh: R*R -> Z -> nat -> R.
Variable unh : R * R -> Z -> nat -> R.
Variable eps:R.

Definition L (u : R * R -> R) (X : R * R) : R :=
  (partial_derive_secondvar_n partial_derive_secondvar 2%nat u) X -
  c ^ 2 * (partial_derive_firstvar_n partial_derive_firstvar 2%nat u) X.

Definition L1  (u: R*R -> R) (X:R*R) : R :=
   (partial_derive_secondvar u) X.

Definition L0  (u: R*R -> R) (X:R*R) : R :=
  u X.


Definition is_solution (u: R*R -> R)   :=
  (forall X:R*R, xmin <= fst X  <= xmax -> L u X=f X) /\ 
  (forall x:R, xmin <= x <= xmax -> L1 u (x,0) = u1 x) /\
  (forall x:R, xmin <= x <= xmax -> L0 u (x,0) = u0 x) /\
  (forall t:R, u(xmin,t) = 0) /\  
  (forall t:R, u(xmax,t) = 0). 


Hypothesis usol_is_sol: is_solution usol.

(** Discretization *)

Definition discretization (u : R * R -> R) (dX : R * R) : Z -> nat-> R :=
  fun j k => u (Xjk dX j k).

Definition ubar (dX : R * R) (j:Z) (k : nat) : R :=
   discretization usol dX j k. 


(** Numerical scheme *)

Definition Ah (unh : Z -> R) (dx : R) (j : Z) : R :=
  - (c *c) * (unh (j + 1)%Z - 2 * unh j + unh (j - 1)%Z) / (dx *dx).



Definition Lh (uh: Z -> nat -> R) (dX : R * R) (j:Z) (n : nat) : R :=
   match n with
     O => uh j 0%nat
   | 1 => (uh j 1%nat - uh j 0%nat)/ (snd dX) + (snd dX) / 2 * (Ah (fun  j' => uh j' 0%nat) (fst dX) j)
   | S (S  m) =>  (uh j (m + 2)%nat - 2 * uh j (m+1)%nat + uh j m) / (snd dX *snd dX) + 
       (Ah (fun j' => uh j' (m+1)%nat) (fst dX) j)
   end.

(** The scheme is defined so that unh will be zero for j <= 0 or j >= ni *)

Definition scheme (uh: R*R -> Z -> nat -> R) (dX : R * R) (j:Z) (n : nat) : R :=
   match (Z_lt_le_dec 0 j) with
   | left _ => match  (Z_lt_le_dec j  (ni (fst dX))) with
         | left  _ => match n with
                 O => Lh (uh dX) dX j 0 - u0h dX j  
               | 1 => Lh (uh dX) dX j 1 - u1h dX j
               | S (S m) => Lh (uh dX) dX j (S (S m)) - fh dX j (S m)
               end
         | right _ => uh dX j n
         end
   | right  _ => uh dX j n
   end.


Definition is_discrete_solution (dX:R*R) := snd dX <> 0 ->
 forall  j n, (0 <= j <= ni (fst dX))%Z -> scheme unh dX j n = 0.


Definition LL1h (v: R*R->R) (X dX:R*R) : R:=
   (v (fst X , snd X+snd dX) - v X)/ (snd dX) - partial_derive_secondvar v X -
   (snd dX) / 2 * (c*c) *(v (fst X +fst dX, snd X) - 2*v X + 
   v (fst X -fst dX, snd X))/(fst dX*fst dX).

(** Energy and errors *)

Definition energy (uh : Z -> nat -> R) (dX : R * R) (n : nat) :=
  1/2 *
    finite_squared_norm_dx (fst dX)
      (scalar_multiply_n (1 / snd dX)
        (subtract_n
          (fun j => uh j (n + 1)%nat)
          (fun j => uh j n))) +
  1/2 *
    finite_dot_product_dx (fst dX)
      (fun j => (Ah (fun j' => uh j' n) (fst dX) j)) 
      (fun j => uh j (n + 1)%nat).


Definition truncation_error := (scheme ubar).


Definition convergence_error (dX : R * R) (j:Z) (n : nat) : R :=
  (ubar dX j n) - (unh dX j n).


(** A few properties, mainly of Ah *)



Lemma jx_Betw: forall (x:{x| Betw xmin xmax x}) dx, 0 < dx ->
   (0 <= jx (projT1 x-xmin) dx <= ni dx)%Z.
intros (x,(Hx1,Hx2)) dx H; simpl; unfold jx.
split.
apply Int_part_pos.
unfold Rdiv; apply Rmult_le_pos.
apply Rplus_le_reg_l with xmin; now ring_simplify.
auto with real.
unfold R_n.ni, jx.
apply Int_part_monotone.
apply Rmult_le_compat_r.
auto with real.
apply Rplus_le_reg_l with xmin; now ring_simplify.
Qed.




Lemma Xjk_Betw: forall (x:{x| Betw xmin xmax x}) dX i, 0 < fst dX ->
   xmin <= fst (Xjk dX (jx (projT1 x - xmin) (fst dX)) i) <= xmax.
intros a dX i H.
specialize (jx_Betw a (fst dX) H).
destruct a as (x,(Hx1,Hx2)).
simpl; unfold jx; intros (H1,H2).
split.
apply Rplus_le_reg_l with (-xmin); ring_simplify.
apply Rmult_le_pos; auto with real zarith.
replace 0%R with (IZR 0); try apply IZR_le; auto with real zarith.
apply Rle_trans with (2:=Hx2).
apply Rplus_le_reg_l with (-xmin).
apply Rmult_le_reg_l with (/ fst dX); auto with real.
apply Rle_trans with (IZR (Int_part ((x - xmin) / fst dX))).
right; field; auto with real.
apply Rle_trans with (1:=proj1 (base_Int_part ((x - xmin) / fst dX))).
right; unfold Rdiv; ring.
Qed.

Lemma Xjk_Betw_j: forall dX j i, 0 < fst dX -> 
   (0 <= j <= ni (fst dX))%Z ->
   xmin <= fst (Xjk dX j i) <= xmax.
intros dX j i H Hj.
simpl; unfold jx; split.
apply Rplus_le_reg_l with (-xmin); ring_simplify.
apply Rmult_le_pos; auto with real zarith.
replace 0%R with (IZR 0); try apply IZR_le; auto with real zarith.
apply Rplus_le_reg_l with (-xmin).
apply Rmult_le_reg_l with (/ fst dX); auto with real.
apply Rle_trans with (IZR j).
right; field; auto with real.
apply Rle_trans with (IZR (ni (fst dX))).
apply IZR_le; auto with zarith.
unfold R_n.ni, jx.
apply Rle_trans with (1:=proj1 (base_Int_part ((xmax - xmin) / fst dX))).
right; unfold Rdiv; ring.
Qed.


Lemma scheme_eq_1: forall uh dX i n, (i <= 0)%Z ->
  scheme uh dX i n =  uh dX i n.
intros; unfold scheme.
case (Z_lt_le_dec 0 i); intros; auto.
contradict H; auto with zarith.
Qed.

Lemma scheme_eq_2: forall uh dX i n, (ni (fst dX) <= i)%Z ->
  scheme uh dX i n =   uh dX i n.
intros; unfold scheme.
case (Z_lt_le_dec 0 i); intros; auto.
case (Z_lt_le_dec i (ni (fst dX))); intros; auto.
contradict H; auto with zarith.
Qed.

Lemma scheme_eq_3: forall uh dX j n, (0 < j < ni (fst dX))%Z ->
  scheme uh dX j n =  match n with
                 O => Lh (uh dX) dX j 0 - u0h dX j  
               | 1 => Lh (uh dX) dX j 1 - u1h dX j
               | S (S m) => Lh (uh dX) dX j (S (S m)) - fh dX j (S m)
               end.
intros; unfold scheme.
case (Z_lt_le_dec 0 j); intros.
case (Z_lt_le_dec j (ni (fst dX))); intros; auto.
contradict l0; auto with zarith.
contradict l; auto with zarith.
Qed.



Lemma Ah_is_linear: forall uh1 uh2 dx j,
  Ah (fun j0 => uh1 j0 - uh2 j0) dx j = Ah uh1 dx j - Ah uh2 dx j.
intros; unfold Ah; unfold Rdiv; ring.
Qed.


Lemma Lh_is_linear: forall uh1 uh2 dX j n,
  Lh (fun j0 n0 => uh1 j0 n0 - uh2 j0 n0) dX j n = Lh uh1 dX j n - Lh uh2 dX j n.
intros; unfold Lh.
case n; auto; intros; case n0.
rewrite Ah_is_linear; unfold Rdiv; ring.
intros; rewrite Ah_is_linear; unfold Rdiv; ring.
Qed.

Lemma Green_formula: forall dx u v,
  0 < dx ->
 v 0%Z = 0 -> v (ni dx) = 0 ->
 finite_dot_product_dx dx (fun n => u (n+1)%Z -2* u n + u (n-1)%Z) v
  = - finite_dot_product_dx dx (fun n => u (n+1)%Z - u n)
                                                     (fun n => v (n+1)%Z - v n).
intros.
rewrite finite_dot_product_dx_eq_l with (xmin:=xmin) (xmax:=xmax) 
  (v:= (add_n (fun j => u (j + 1)%Z ) 
        (add_n (scalar_multiply_n (-2) (fun j => u j))
              (fun j => u (j - 1)%Z )))); auto.
2: unfold scalar_multiply_n, add_n; intros; ring.
rewrite finite_dot_product_dx_eq_l with 
  (u:=(fun n : Z => u (n + 1)%Z - u n)) 
  (v:=subtract_n (fun n => u (n+1)%Z) u); auto.
rewrite finite_dot_product_dx_eq_r with 
  (u:=(fun n : Z => v (n + 1)%Z - v n)) 
  (v:=subtract_n (fun n => v (n+1)%Z) v); auto.
repeat rewrite finite_dot_product_dx_add_l.
rewrite finite_dot_product_dx_sub_l.
repeat rewrite finite_dot_product_dx_sub_r.
rewrite finite_dot_product_dx_scal_mult_l.
rewrite <- finite_dot_product_dx_shift_indices; auto.
rewrite finite_dot_product_dx_shift_indices with 
  (f1:=(fun j : Z => u (j - 1)%Z)); auto.
rewrite finite_dot_product_dx_eq_l with 
  (v:=u) (u:=(fun j : Z => u (j + 1 - 1)%Z)); auto.
2: intros; apply f_equal; auto with zarith.
rewrite finite_dot_product_dx_eq_l with 
  (v:=u) (u:=(fun j : Z => u j)); auto.
ring.
Qed.

Lemma Ah_is_symmetrical :
  forall dx u v,
  0 < dx -> 
  u 0%Z = 0 -> u (ni dx) = 0 -> v 0%Z = 0 -> v (ni dx) = 0 ->
    finite_dot_product_dx dx (Ah u dx) v
   = finite_dot_product_dx dx u (Ah v dx).
intros; unfold Ah.
rewrite finite_dot_product_dx_eq_l with  (v:=(scalar_multiply_n (-c*c/(dx*dx))
   (fun n => u (n + 1)%Z - 2 * u n + u (n - 1)%Z))); auto.
2: unfold scalar_multiply_n, Rdiv; intros; ring.
rewrite finite_dot_product_dx_eq_r with  (v:=(scalar_multiply_n (-c*c/(dx*dx))
   (fun n => v (n + 1)%Z - 2 * v n + v (n - 1)%Z))) (w:=u); auto.
2: unfold scalar_multiply_n, Rdiv; intros; ring.
rewrite finite_dot_product_dx_sym with (u:=u).
repeat rewrite finite_dot_product_dx_scal_mult_l.
apply f_equal.
repeat rewrite Green_formula; auto.
rewrite finite_dot_product_dx_sym; auto with real.
Qed.

Lemma Ah_eq: forall uh1 uh2 dx, (forall j, (0 <= j <= ni dx)%Z -> uh1  j=uh2  j) -> 
  forall j, (0 < j < ni dx)%Z -> Ah uh1 dx j= Ah uh2 dx j.
intros;unfold Ah.
repeat rewrite H; auto with zarith.
Save.


Lemma Ah_cont: forall dx vh,
  0 < dx ->
  vh 0%Z = 0 -> vh (ni dx) = 0 ->
  finite_dot_product_dx dx  (Ah vh dx) vh 
      <= 4*c*c/(dx*dx)*finite_squared_norm_dx dx vh. 
intros; unfold Ah.
rewrite finite_dot_product_dx_eq_l with  (v:=(scalar_multiply_n ((c*c/(dx*dx))*(-1))
   (fun n => vh (n + 1)%Z - 2 * vh n + vh (n - 1)%Z))); auto.
2: unfold scalar_multiply_n, Rdiv; intros; ring.
rewrite finite_dot_product_dx_scal_mult_l.
rewrite Rmult_assoc.
apply Rle_trans with
  ((c * c / (dx * dx))*(4 * finite_squared_norm_dx dx vh));[idtac|right; unfold Rdiv; ring].
apply Rmult_le_compat_l.
apply Rmult_le_pos.
fold (Rsqr c); auto with real.
left; apply Rinv_0_lt_compat; auto with real.
fold (Rsqr dx); auto with real.
rewrite Green_formula; auto.
ring_simplify.
fold (R_n.finite_squared_norm_dx xmin xmax dx  (fun n : Z => vh (n + 1)%Z - vh n)).
apply Rle_trans with (2 * 
  (finite_squared_norm_dx dx (fun n => vh (n+1)%Z) + finite_squared_norm_dx dx vh)).
apply finite_squared_norm_parall_sub.
auto with real.
replace (finite_squared_norm_dx dx (fun n : Z => vh (n + 1)%Z)) with
    (finite_squared_norm_dx dx vh).
right; ring.
unfold R_n.finite_squared_norm_dx.
rewrite <- finite_dot_product_dx_shift_indices; auto.
Qed.

Lemma Ah_pos: forall dx vh,
  0 < dx -> vh 0%Z = 0 -> vh (ni dx) = 0 ->
  0 <= finite_dot_product_dx dx  (Ah vh dx) vh.
intros; unfold Ah.
rewrite finite_dot_product_dx_eq_l with  (v:=(scalar_multiply_n (-c*c/(dx*dx))
   (fun n => vh (n + 1)%Z - 2 * vh n + vh (n - 1)%Z))); auto.
2: unfold scalar_multiply_n, Rdiv; intros; ring.
rewrite finite_dot_product_dx_scal_mult_l.
rewrite Green_formula; auto.
rewrite Ropp_mult_distr_r_reverse.
unfold Rdiv; repeat rewrite Ropp_mult_distr_l_reverse.
rewrite Ropp_involutive.
apply Rmult_le_pos.
apply Rmult_le_pos.
fold (Rsqr c); auto with real.
left; apply Rinv_0_lt_compat.
apply Rmult_lt_0_compat; auto with real.
apply finite_squared_norm_dx_pos.
left; trivial.
Qed.


(** Scheme properties *)

Hypothesis eps_lt: 0 < eps < 1.

Definition CFL (dX:R*R)  := c * (snd dX) / (fst dX).

Definition CFL_condition_r (dX:R*R) := CFL dX <=  1 - eps.

Definition ni_exact (dX:R*R) := (IZR (ni (fst dX)) = (xmax-xmin) / (fst dX))%R.

Notation Oupse := (OuP (fun dX => 0 < fst dX /\ 0 < snd dX /\ ni_exact dX)).

Notation OupsCFLe := (OuP (fun dX => 0 < fst dX /\ 0 < snd dX /\ CFL_condition_r dX /\ ni_exact dX)).


Definition scheme_is_consistent_of_order (p q : nat) :=    
    Oupse (fun (t: {t| Under T t}) dX => finite_squared_norm_dx (fst dX)
                          (fun j => truncation_error dX j  (kt (projT1 t) (snd dX))))
      (fun dX => Rsqr ((fst dX) ^ p + (snd dX) ^ q)).

Definition scheme_is_convergent_of_order (p q : nat) :=    
    OupsCFLe (fun (t: {t| Under T t}) dX => finite_norm_dx (fst dX)
                          (fun j => convergence_error dX j  (kt (projT1 t) (snd dX))))
      (fun dX =>  (fst dX) ^ p + (snd dX) ^ q).

End Wave.

(** The discretization is 0 at 0 and ni *)


Section zeroness.

Variable xmin xmax:R.
Hypothesis xminLtxmax: xmin < xmax.
Variable c:R.
Variable dX:R*R.
Variable u0h u1h: R*R -> Z-> R.
Variable fh: R*R -> Z -> nat -> R.
Variable unh : R * R -> Z -> nat -> R.


Hypothesis unh_is_discrete_sol: is_discrete_solution xmin xmax c u0h u1h fh unh dX.

Notation ni := (fun x => (ni xmin xmax x)).


Lemma zeroness_unh_1: forall (n:nat), 0 < fst dX -> snd dX <> 0 ->
    unh dX 0 n = 0.
intros.
elim (unh_is_discrete_sol H0 0%Z n);unfold scheme.
case (Z_lt_le_dec 0 0).
intros H'; contradict H'; auto with zarith.
now intros.
split; auto with zarith.
now apply ni_pos.
Qed. 


Lemma zeroness_unh_2: forall (n:nat), 0 < fst dX -> snd dX <> 0 ->
    unh dX (ni (fst dX)) n = 0.
intros.
elim (unh_is_discrete_sol H0 (ni (fst dX)) n);unfold scheme.
case (Z_lt_le_dec 0 (ni (fst dX))).
case (Z_lt_le_dec (ni (fst dX)) (ni (fst dX))).
intros H'; contradict H'; auto with zarith.
now intros.
now intros.
split; auto with zarith.
now apply ni_pos.
Qed.


End zeroness.



Section zeroness2.

Variable xmin xmax:R.
Hypothesis xminLtxmax: xmin < xmax.

Variable partial_derive_firstvar: (R * R -> R) -> (R * R -> R).
Variable partial_derive_secondvar: (R * R -> R) -> (R * R -> R).

Variable c:R.
Variable u0 u1: R->R.
Variable f : R*R -> R.
Variable dX:R*R.
Variable usol : R * R -> R.

Hypothesis usol_is_sol: is_solution xmin xmax partial_derive_firstvar partial_derive_secondvar c u0 u1 f usol.

Notation ni x := (ni xmin xmax x).

(** The solution is 0 at 0 and ni as ni exactly corresponds to xmax *)

Lemma zeroness_usol_xmin_ni : forall t, fst dX <> 0 -> ni_exact xmin xmax dX ->
 usol (xmin + IZR (ni (fst dX)) * fst dX, t) =0.
intros.
replace (xmin + IZR (ni (fst dX)) * fst dX) with xmax.
apply usol_is_sol.
rewrite H0; field; assumption.
Qed.


Lemma zeroness_ubar_1: forall (n:nat),
    ubar xmin usol dX 0 n = 0.
intros.
unfold ubar, discretization, Xjk.
rewrite Rmult_0_l, Rplus_0_r.
apply usol_is_sol.
Qed.


Lemma zeroness_ubar_2: forall (n:nat),
    fst dX <> 0 -> ni_exact xmin xmax dX -> 
    ubar xmin usol dX (ni (fst dX)) n = 0.
intros.
unfold ubar, discretization, Xjk.
now apply zeroness_usol_xmin_ni.
Qed.

End zeroness2.

Section my_unh.

Variable xmin xmax c:R.
(* Hypothesis xminLtxmax: xmin < xmax.*)
Variable u0h u1h: R*R -> Z-> R.
Variable fh: R*R -> Z -> nat -> R.


Fixpoint my_unh (dX:R*R) (j:Z) (n:nat) {struct n} : R := 
   if Z_lt_le_dec 0 j then
   if Z_lt_le_dec j (ni xmin xmax (fst dX)) then
    match n with
     | 0%nat => u0h dX j
     | S (0%nat as n1) => my_unh dX j n1 - snd dX*(snd dX/2*Ah c (fun j'=> my_unh dX j' n1) (fst dX)j - u1h dX j)
     | S (S n2 as n1)  => (2*my_unh dX j n1 - my_unh dX j n2) 
         - snd dX*snd dX*(Ah c (fun j'=> my_unh dX j' n1) (fst dX) j -fh dX j n1)
     end
  else 0%R
  else 0%R.


Lemma my_unh_correct: forall dX, is_discrete_solution xmin xmax c u0h u1h fh my_unh dX.
intros dX H0 j n Hj.
unfold scheme.
case (Z_lt_le_dec 0 j).
case (Z_lt_le_dec j (ni xmin xmax (fst dX))).
(* *)
intros Hj1 Hj2.
case n.
(* . *)
simpl.
case (Z_lt_le_dec 0 j); [idtac|intros V; contradict V; auto with zarith].
case (Z_lt_le_dec j (ni xmin xmax (fst dX))); [idtac|intros V; contradict V; auto with zarith].
intros _ _; ring.
(* . *)
clear n; intros n; case n.
simpl.
case (Z_lt_le_dec 0 j); [idtac|intros V; contradict V; auto with zarith].
case (Z_lt_le_dec j (ni xmin xmax (fst dX))); [idtac|intros V; contradict V; auto with zarith].
intros _ _.
field; auto with real.
(* . *)
clear n; intros n.
unfold Lh.
replace (n+2)%nat with (S (S n)) by ring.
replace (my_unh dX j (S (S n))) with
    ((2*my_unh dX j (S n) - my_unh dX j n) 
         - snd dX*snd dX*(Ah c (fun j'=> my_unh dX j' (S n)) (fst dX) j -fh dX j (S n))).
replace (n+1)%nat with (S n) by ring.
field; auto.
simpl.
case (Z_lt_le_dec 0 j); [idtac|intros V; contradict V; auto with zarith].
case (Z_lt_le_dec j (ni xmin xmax (fst dX))); [auto|intros V; contradict V; auto with zarith].
(* *)
intros Hj2; replace j with (ni xmin xmax (fst dX)); auto with zarith.
intros; case n; simpl; auto.
case (Z_lt_le_dec 0 (ni xmin xmax (fst dX))); auto with zarith.
case (Z_lt_le_dec (ni xmin xmax (fst dX)) (ni xmin xmax (fst dX))); auto with zarith.
intros V; contradict V; auto with zarith.
intros; case (Z_lt_le_dec 0 (ni xmin xmax (fst dX))); auto with zarith.
case (Z_lt_le_dec (ni xmin xmax (fst dX)) (ni xmin xmax (fst dX))); auto with zarith.
intros V; contradict V; auto with zarith.
intros; replace j with 0%Z; auto with zarith.
case n; simpl; auto.
Qed.



Lemma scheme_eq: forall unh1 unh2 nmax dX, snd dX <> 0 ->
    (forall  j n, (0 <= j <= ni xmin xmax (fst dX))%Z -> (n <= nmax)%nat -> 
         scheme xmin xmax c u0h u1h fh unh1 dX j n = scheme xmin xmax c u0h u1h fh unh2 dX j n)
 -> (forall  j n, (0 <= j <= ni xmin xmax (fst dX))%Z -> (n <= nmax)%nat -> unh1 dX j n = unh2 dX j n).
intros unh1 unh2 nmax dX H1 H j n; generalize j; clear j.
apply lt_wf_ind with (P:=fun n => forall j : Z, (0 <= j <= ni xmin xmax (fst dX))%Z -> (n <= nmax)%nat 
   -> unh1 dX j n = unh2 dX j n).
clear n; intros n IHn j Hj Hn.
(* *)
case_eq n.
intros Hn1; rewrite Hn1 in Hn.
generalize (H j 0%nat Hj Hn); unfold scheme.
case (Z_lt_le_dec 0 j); auto.
case (Z_lt_le_dec j (ni xmin xmax (fst dX))); auto.
unfold Lh; intros; auto with real.
apply trans_eq with ((unh1 dX j 0%nat - u0h dX j) + u0h dX j).
ring.
rewrite H0; ring.
(* *)
intros n1 Hnn1.
case_eq n1.
intros Hn1; rewrite Hn1 in Hnn1; rewrite Hnn1 in Hn.
generalize (H j 1%nat Hj Hn); unfold scheme.
case (Z_lt_le_dec 0 j); auto.
case (Z_lt_le_dec j (ni xmin xmax (fst dX))); auto.
unfold Lh; intros; auto with real.
apply Rmult_eq_reg_l with (/snd dX); auto with real.
apply Rplus_eq_reg_l with ((- unh1 dX j 0%nat) / snd dX +
     snd dX / 2 * Ah c (fun j' : Z => unh1 dX j' 0%nat) (fst dX) j - 
     u1h dX j).
apply trans_eq with ((unh1 dX j 1%nat - unh1 dX j 0%nat) / snd dX +
     snd dX / 2 * Ah c (fun j' : Z => unh1 dX j' 0%nat) (fst dX) j - 
     u1h dX j);[unfold Rdiv; ring|idtac].
rewrite H0.
unfold Ah; repeat rewrite <- (IHn 0%nat); auto with zarith.
unfold Rdiv; ring.
(* *)
intros n2 Hn1n2.
rewrite Hn1n2 in Hnn1; rewrite Hnn1 in Hn.
generalize (H j (S (S n2)) Hj Hn); unfold scheme.
case (Z_lt_le_dec 0 j); auto.
case (Z_lt_le_dec j (ni xmin xmax (fst dX))); auto.
unfold Lh; intros; auto with real.
replace (S (S n2)) with (n2+2)%nat; [idtac|ring].
apply Rmult_eq_reg_l with (/(snd dX*snd dX)); auto with real.
apply Rplus_eq_reg_l with ((- 2 * unh1 dX j (n2 + 1)%nat + unh1 dX j n2) /
     (snd dX * snd dX) +
     Ah c (fun j' : Z => unh1 dX j' (n2 + 1)%nat) (fst dX) j - fh dX j (S n2)).
apply trans_eq with ((unh1 dX j (n2 + 2)%nat - 2 * unh1 dX j (n2 + 1)%nat + unh1 dX j n2) /
     (snd dX * snd dX) +
     Ah c (fun j' : Z => unh1 dX j' (n2 + 1)%nat) (fst dX) j - fh dX j (S n2));[unfold Rdiv; ring|idtac].
rewrite H0; unfold Ah.
repeat rewrite <- (IHn (n2+1)%nat); auto with zarith.
rewrite <- (IHn n2); auto with zarith.
unfold Rdiv; ring.
Qed.

End my_unh.













